unit ICWindows;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module represents the external interface to the IC window management
	system, providing routines to open and close windows, handle window related
	events, and save and restore window positions.
	
	The core data structures for this module are made public in ICWindowGlobals
	because they need to be accessed by each of the Whats. That module contains
	a lot more comments on how IC window management works.
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Menu Handling ***** *)
	
	procedure WindowsAdjustEditMenu;
		(* The routine adjusts the Edit menu to reflect the current window state. *)
	
	procedure WindowsDoEditMenu (item: integer);
		(* This routine dispatches a menu command from the Edit menu. *)

	(* ***** Basic Event Handling ***** *)
	
	function WindowsEarlyHandleKey (event: EventRecord): Boolean;
		(* This routine is called by ICEvents after command key processing
			but before the event is sent to the Dialog Manager. This allows
			the window to process keyboard events before the Dialog Manager
			sees them. The only use of this at the moment is to handle
			tab key presses in a special way.
		*)
	
	procedure WindowsDoKey (event: EventRecord);
		(* This routine is called by ICEvents in response to a keyboard event
			that hasn't been processed by the Dialog Manager.
			The main action is to pass these events along to the What for item
			dependent processing.
		*)
	
	procedure WindowsIdle;
		(* This routine is periodically called by the main event loop. It's used
			for two purposes. The first is to blink the caret in the currently
			selected text field. The second is to track the cursor over the various
			dialog items and set the cursor appropriately.
		*)

	procedure WindowActivateDeactivate (window: WindowPtr; activate: Boolean);
		(* This routine is called by ICEvents in response to a activate or
			deactivate event. The main action is to pass these events along to the
			What for item dependent processing. It also special cases the activation
			of the currently selected text What.
		*)
	
	procedure WindowItemClick (window: WindowPtr; event: EventRecord; item: integer);
		(* This routine is called by ICEvents in response to a click on
			an active dialog item
			The main action is to pass these events along to the What for item
			dependent processing.
		*)

	(* ***** Opening and Closing ***** *)

	function WindowOpen (wt: WindowType): OSStatus;
		(* This routine opens the window specified by wt. *)
	
	function WindowClose (wt: WindowType): OSStatus;
		(* This routine closes the window specified by wt. *)
	
	procedure WindowSetTitle (wt: WindowType; title: Str255);
		(* This routine changes the window title of the window. *)

	function WindowsFlushAll: OSStatus;
		(* This routine flushes (ie saves) all information windows. By information
			window I mean windows that are showing user preferences. Basically this
			is all the windows except the about box and the main window.
		*)

	function WindowsCloseAll: OSStatus;
		(* This routine closes all information windows. By information window
			I mean windows that are showing user preferences. Basically this
			is all the windows except the about box and the main window.
			Unsaved changes are automatically flushed.
		*)

	(* ***** Window Positioning ***** *)

	procedure WindowsResetPositions;
		(* This routine resets position of all the windows to their defaults. *)
	
	procedure WindowsSavePositions;
		(* This routine saves the position of all the windows into the
			current preferences file.
		*)
	
	procedure WindowsRestorePositions;
		(* This routine restores the position of all the windows from the
			current preferences file.
		*)

	(* ***** Misc Stuff ***** *)
	
	function InitICWindows: OSStatus;
		(* This routine initialises the module. *)
	
implementation

	uses
		Icons,
		QuickDraw, 
		Resources, 
		ToolUtils, 
		Dialogs,

		ICTypes, 
		ICAPI, 
		ICKeys,

		ICResources,		
		ICStringLists, 
		ICMiscSubs, 
		ICDialogs, 
		ICText, 

		ICAboutWhats,
		ICTextWhats,

		ICGlobals, 
		ICDocument;

	const
		kOurWindowPositionsKey = '49434170WindowPositions';

	(* ***** Menu Handling ***** *)

	procedure WindowsAdjustEditMenu;
		(* See comment in interface part. *)
	begin
		if MapWindowPtrToWindowType(FrontWindow) = WT_About then begin
			AdjustAboutMenu;
		end else begin
			AdjustTextMenu(MapWindowPtrToWindowType(FrontWindow));
		end; (* if *)
	end; (* WindowsAdjustEditMenu *)

	procedure WindowsDoEditMenu (item: integer);
		(* See comment in interface part. *)
		var
			wt: WindowType;
	begin
		wt := MapWindowPtrToWindowType(FrontWindow);
		case wt of
			WT_About :
				DoAboutMenu(item);
			WT_None :
				(* do nothing *) ;
			otherwise
				DoTextMenu(wt, item);
		end; (* case *)
	end; (* WindowsDoEditMenu *)

	(* ***** Event Handling Utilities ***** *)

	procedure WindowTab (wt : WindowType; shift: Boolean);
		(* Handle a tab in the specified window. Tabs forwards or backward depending
			on the shift Boolean.
		*)
		var
			originalItem: integer;
			increment: integer;
			itemToSelect: integer;
			currentItem : integer;
			what : WhatRecordPtr;
	begin
		(* Set up the place to start the search. Start the search from the currently
			selected item, or at the start/end.
		*)
		originalItem := GetSelectedTextItem(wt);
		if originalItem <= 0 then begin
			if shift then begin
				originalItem := 1;
			end else begin
				originalItem := kMaxItemsPerWindow;
			end;
		end;
		
		(* Calculate the increment. Because we use modulo arithmetic in the loop
			the increment is either set to kMaxItemsPerWindow or kMaxItemsPerWindow - 2,
			which translates to +1 or -1 when you take the mod with kMaxItemsPerWindow
			and then add 1 because we're using a 1 based array.
		*)
		increment := kMaxItemsPerWindow - 2 * ord(shift);

		(* Now search the window for a TEXT item to tab to. *)
		itemToSelect := -1;
		currentItem := originalItem;
		repeat
			currentItem := (currentItem + increment) mod kMaxItemsPerWindow + 1;
			what := gWindows[wt].items[currentItem];
			if (what <> nil) & (what^.whatClass^.whatTypeID = 'TEXT') then begin
				itemToSelect := currentItem;
			end; (* if *)
		until (currentItem = originalItem) or (itemToSelect <> -1);
		
		(* Now select the item we found. *)
		if itemToSelect > 0 then begin
			SelectTextItem(wt, itemToSelect);
		end; (* if *)
	end; (* WindowTab *)

	(* ***** Basic Event Handling ***** *)

	function WindowsEarlyHandleKey (event: EventRecord): Boolean;
		(* See comment in interface part. *)
		var
			result: Boolean;
			ch: char;
			wt: WindowType;
	begin
		result := false;
		wt := MapWindowPtrToWindowType(FrontWindow);
		if wt <> WT_None then begin
			if not gWindows[wt].sendTabsToWhats then begin
				ch := chr(band(event.message, charCodeMask));
				if ch = kTabChar then begin
					WindowTab(wt, band(event.modifiers, shiftKey) <> 0);
					result := true;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		WindowsEarlyHandleKey := result;
	end; (* WindowsEarlyHandleKey *)

	procedure WindowsDoKey (event: EventRecord);
		(* See comment in interface part. *)
		var
			wt: WindowType;
			selectedItem: integer;
			what : WhatRecordPtr;
			keyProc: WhatKeyProcPtr;
	begin
		wt := MapWindowPtrToWindowType(FrontWindow);
		if wt <> WT_None then begin
			selectedItem := gWindows[wt].selectedItem;
			if selectedItem > 0 then begin
				keyProc := nil;
				what := gWindows[wt].items[selectedItem];
				if what <> nil then begin
					keyProc := what^.whatClass^.keyProc;
				end; (* if *)
				if keyProc <> nil then begin
					SetPort(gWindows[wt].window);
					DisplayError(acDoThis, keyProc(wt, selectedItem, event));
				end; (* if *)
			end; (* if *)
		end; (* if *)
	end; (* WindowsDoKey *)

	procedure WindowsIdle;
		(* See comment in interface part. *)
		var
			wt: WindowType;
			selectedItem: integer;
			mouseLocation: Point;
			cursorSet: Boolean;
			anItem : integer;
			theItemRect: Rect;
			what: WhatRecordPtr;
			cursorid: integer;
			cursorProc: WhatCursorProcPtr;
			junk: OSStatus;
	begin
		wt := MapWindowPtrToWindowType(FrontWindow);
		if wt <> WT_None then begin
			SetPort(gWindows[wt].window);
			
			(* If we have a selected text item, idle the text associated with it. *)
			selectedItem := GetSelectedTextItem(wt);
			if selectedItem > 0 then begin
				TextIdle(gWindows[wt].items[selectedItem]^.whatData);
			end; (* if *)
			
			if (gWindows[wt].window = FrontWindow) & InForeground then begin

				(* Now, if we're at the front, do the cursor tracking stuff. *)
				GetMouse(mouseLocation);
				cursorSet := false;
				for anItem := 1 to kMaxItemsPerWindow do begin
					what := gWindows[wt].items[anItem];
					if what <> nil then begin
						GetDItemRect(gWindows[wt].window, anItem, theItemRect);
						if PtInRect(mouseLocation, theItemRect) then begin
							(* The mouse pointer is inside this item, set the cursor
								based on the cursorProc and cursor ID stored in the
								class record.
							*)
							cursorSet := true;
							cursorid := what^.whatClass^.cursorID;
							cursorProc := what^.whatClass^.cursorProc;
							if cursorProc = nil then begin
								if cursorid = 0 then begin
									InitCursor;
								end else begin
									SetCursor(GetCursor(cursorid)^^);
								end; (* if *)
							end else begin
								junk := cursorProc(wt, anItem, mouseLocation, cursorid);
							end; (* if *)
						end; (* if *)
					end; (* if *)
				end; (* for *)
				
				(* If we haven't set the cursor already, set it back to the pointer. *)
				if not cursorSet then begin
					InitCursor;
				end; (* if *)
			end; (* if *)
		end; (* if *)
	end; (* WindowsIdle *)

	procedure WindowActivateDeactivate (window: WindowPtr; activate: Boolean);
		(* See comment in interface part. *)
		var
			wt: WindowType;
			selectedItem: integer;
			anItem: integer;
			what: WhatRecordPtr;
			activateProc: WhatActivateProcPtr;
			junk: OSStatus;
	begin
		wt := MapWindowPtrToWindowType(window);
		if wt <> WT_None then begin
			SetPort(window);
			
			(* If we have a selected text item, activate/deactivate the text associated with it. *)
			selectedItem := GetSelectedTextItem(wt);
			if selectedItem > 0 then begin
				TextActivate(gWindows[wt].items[selectedItem]^.whatData, activate);
			end; (* if *)

			(* Now call each What in the window which is interested in activate/deactivate. *)
			for anItem := 1 to kMaxItemsPerWindow do begin
				what := gWindows[wt].items[anItem];
				if what <> nil then begin
					activateProc := what^.whatClass^.activateProc;
					if activateProc <> nil then begin
						junk := activateProc(wt, anItem, activate);
					end; (* if *)
				end; (* if *)
			end; (* for *)
		end; (* if *)
	end; (* WindowActivateDeactivate *)
	
	procedure WindowItemClick (window: WindowPtr; event: EventRecord; item: integer);
		(* See comment in interface part. *)
		var
			wt: WindowType;
			clickProc: WhatClickProcPtr;
			what : WhatRecordPtr;
	begin
		wt := MapWindowPtrToWindowType(window);
		clickProc := nil;
		what := gWindows[wt].items[item];
		if what <> nil then begin
			clickProc := what^.whatClass^.clickProc;
		end; (* if *)
		if clickProc <> nil then begin
			SetPort(window);
			DisplayError(acDoThis, clickProc(wt, item, event));
		end; (* if *)
	end; (* WindowItemClick *)

	(* ***** Closing and Saving Internals ***** *)

	function DisposeWindowType (wt: WindowType): OSStatus;
		(* Dispose of the window associated with this the specified window type.
			This is called by the code that closes windows (as you would expect),
			but also by the code that creates windows when the creation fails.
			The routine is expected to deal with this.
		*)
		var
			err: OSStatus;
			firstErr : OSStatus;
			anItem: integer;
			closeProc: WhatCloseProcPtr;
			what: WhatRecordPtr;
	begin
		firstErr := noErr;
		SetPort(gWindows[wt].window);
		
		(* Iterate through each item, disposing it if necessary. *)
		for anItem := 1 to kMaxItemsPerWindow do begin
			what := gWindows[wt].items[anItem];
			if what <> nil then begin
				closeProc := what^.whatClass^.closeProc;
				if closeProc <> nil then begin
					err := closeProc(wt, anItem);
					if firstErr = noErr then begin
						firstErr := err;
					end; (* if *)
				end; (* if *)
				gWindows[wt].items[anItem] := nil;
				DisposePtr(Ptr(what));
			end; (* if *)
		end; (* for *)

		(* Dispose the actual window. *)
		DisposeDialog(gWindows[wt].window);
		gWindows[wt].window := nil;
		
		DisposeWindowType := firstErr;
	end; (* DisposeWindowType *)

	function FlushWindowType (wt: WindowType): OSStatus;
		(* This code flushes all the items associated with the designated window type.
			It is called by the code that closes windows, and by the WindowsFlushAll
			routine.

			The only (minor) complication is that it records the window position in the
			global array of windows, so that the values are up-to-date.
			This means that the WindowsSavePositions code can just save this array.
			This routine is useful for this because it's called prior to window saves
			and closes. The WindowsSavePositions can't do this because the windows
			are already closed by the time it gets called.
			
			Note that this routine does not begin an IC session. This is because it's
			only called from CloseWindowType and WindowsFlushAll, both of which
			begin the session prior to calling this routine.
		*)
		var
			firstErr : OSStatus;
			err: OSStatus;
			portRect: Rect;
			anItem: integer;
			what: WhatRecordPtr;
			flushProc: WhatFlushProcPtr;
	begin
		firstErr := noErr;
		SetPort(gWindows[wt].window);
		
		(* Save away the window position. *)
		GetWindowRect(gWindows[wt].window, portRect);
		gWindows[wt].position := portRect.topLeft;
		
		(* Now flush each item in the window. *)
		for anItem := 1 to kMaxItemsPerWindow do begin
			what := gWindows[wt].items[anItem];
			if what <> nil then begin
				flushProc := what^.whatClass^.flushProc;
				if flushProc <> nil then begin
					err := flushProc(wt, anItem);
					if firstErr = noErr then begin
						firstErr := err;
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* for *)
		
		FlushWindowType := firstErr;
	end; (* FlushWindowType *)

	function CloseWindowType (wt: WindowType): OSStatus;
		(* This routine closes the window associated with the supplied window type.
			It performs two basic functions: saving the data associated with the window
			and disposing of the window. These functions are actually performed by
			FlushWindowType and DisposeWindowType. This is just a convenient wrapper
			function.
			
			It also begins an IC session before calling FlushWindowType. This helps
			speed up the flush.
		*)
		var
			err: OSStatus;
			err2: OSStatus;
			opened: Boolean;
	begin
		(* Begin an IC session. *)
		opened := false;
		err := ICBegin(GetDocumentInstance, icReadWritePerm);
		
		(* Flush the window data. *)
		if err = noErr then begin
			opened := true;
			err := FlushWindowType(wt);
		end; (* if *)
		
		(* Kill off the window. *)
		err2 := DisposeWindowType(wt);
		
		(* Clean up. *)
		if err = noErr then begin
			err := err2;
		end; (* if *)
		if opened then begin
			err2 := ICEnd(GetDocumentInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		CloseWindowType := err;
	end; (* CloseWindowType *)

	(* ***** Newing and Opening Internals ***** *)

	procedure ClearWindowItems(wt : WindowType);
		(* This routine clears the item list associated with a window. It is needed because
			you must clear the items before you can start creating the window, just
			in case you get an error and have to call DisposeWindowType. That routine
			expects that any non-nil item pointers are real pointers to Whats and 
			disposes them.
		*)
		var
			anItem : integer;
	begin
		for anItem := 1 to kMaxItemsPerWindow do begin
			gWindows[wt].items[anItem] := nil;
		end; (* for *)
	end; (* ClearWindowItems *)

	function ParseWhatTemplate (wt: WindowType): OSStatus;
		(* This routine creates the What data structure associated with the window type
			from a What template resource.
		*)
		var
			err: OSStatus;
			whatTemplate: WhatTemplateArrayHandle;
			anItem: integer;
			aClass : integer;
	begin
		err := noErr;
		whatTemplate := WhatTemplateArrayHandle(GetResource('WHAT', MapWindowTypeToResourceID(wt)));
		if whatTemplate <> nil then begin
			HLock(Handle(whatTemplate));

			(* Iterate through each item in the template. *)
			for anItem := 1 to GetHandleSize(Handle(whatTemplate)) div sizeof(WhatTemplateRecord) do begin
				if whatTemplate^^[anItem].templateTypeID <> 'NULL' then begin

					(* Create and init the What item associated with this template item. *)
					gWindows[wt].items[anItem] := WhatRecordPtr(NewPtrClear(sizeof(WhatRecord)));
					err := MemError;
					if err <> noErr then begin
						leave;		(* Damnit, can't be bothered getting rid of this leave. *)
					end; (* if *)
					gWindows[wt].items[anItem]^.whatKey := whatTemplate^^[anItem].templateKey;
					gWindows[wt].items[anItem]^.whatFlags := whatTemplate^^[anItem].templateFlags;
					gWindows[wt].items[anItem]^.whatClass := nil;
					
					(* Now setup the whatClass pointer for the item by matching the whatTypeID. *)
					for aClass := 1 to kMaxClass do begin
						if gWhatClasses[aClass].whatTypeID = whatTemplate^^[anItem].templateTypeID then begin
							gWindows[wt].items[anItem]^.whatClass := @gWhatClasses[aClass];
						end; (* if *)
					end; (* for *)
				end; (* if *)
			end; (* for *)
			
			HUnlock(Handle(whatTemplate));
		end; (* if *)
		ParseWhatTemplate := err;
	end; (* ParseWhatTemplate *)
	
	function PrepareWindowWhats (wt: WindowType): OSStatus;
		(* Prepare the Whats associated with the specified window type.
			The strategy is to parse the What template associated with this
			window and create the What's based on that template. Then
			we cycle through the Whats and call their openProcs.
			
			Note that this routine does not begin an IC session. If the What's
			associated with this window expect to have a session already started,
			you have to start it before calling this routine.
			
			There are also some subtle interactions between this routine and
			the text handling. This routine clears the sendTabsToWhats if it
			detects any TEXT Whats in the window. It also sets the 
			selectedItem to -1 before calling the open routines and, if it's
			still -1 at the end of the routine, it calls the tab routine to
			select the first text item. This allows Whats to stop it calling
			WindowTab by setting selectedItem in their open routine.
			This capability is used by the Helper and File Map Whats.
		*)
		var
			err: OSStatus;
			firstErr: OSStatus;
			anItem: integer;
			openProc : WhatOpenProcPtr;
	begin
		(* Do some simple initialisation. It's important to remember that you have to set
			up gWindows[wt].window before calling this routine.
		*)
		SetPort(gWindows[wt].window);
		gWindows[wt].selectedItem := -1;
		gWindows[wt].sendTabsToWhats := true;
		
		(* Now parse the Whats for this window. *)
		firstErr := ParseWhatTemplate(wt);
		
		(* Now iterate through the Whats, calling their openProcs. *)
		if firstErr = noErr then begin
			for anItem := 1 to kMaxItemsPerWindow do begin
				if gWindows[wt].items[anItem] <> nil then begin
					if gWindows[wt].items[anItem]^.whatClass^.whatTypeID = 'TEXT' then begin
						gWindows[wt].sendTabsToWhats := false;
					end; (* if *)
					openProc := gWindows[wt].items[anItem]^.whatClass^.openProc;
					if openProc <> nil then begin
						err := openProc(wt, anItem);
						if firstErr = noErr then begin
							firstErr := err;
						end; (* if *)
					end; (* if *)
				end; (* if *)
			end; (* for *)
		end; (* if *)
		
		(* Now select the first TEXT item. *)
		if firstErr = noErr then begin
			if gWindows[wt].selectedItem = -1 then begin
				WindowTab(wt, false);
			end; (* if *)
		end; (* if *)
		
		PrepareWindowWhats := firstErr;
	end; (* PrepareWindowWhats *)

	function CreateNewWindow (wt: WindowType): OSStatus;
		(* This is the primary mechanism for creating windows. All windows, except
			the about box, are created using this routine.
		*)
		var
			err: OSStatus;
			err2: OSStatus;
			junk: OSStatus;
			savedPosition: Point;
			originalPosition: Rect;
			dialogID: integer;
	begin
		(* Prepare for failure. If this routine fails in any way, it ends up disposing of the
			window using DisposeWindowType. DisposeWindowType expects the window's
			item pointers to be valid or nil, so we have to nil them out here.
		*)
		gWindows[wt].window := nil;
		ClearWindowItems(wt);

		(* Begin a session with IC, which will make the various openProc's
			called by PrepareWindowWhats run a lot faster.  You might wonder why
			we ask for read/write permissions here.  The reason is that
			some of the Whats might want to correct broken preference
			settings as they are opened.
		*)
		err := ICBegin(GetDocumentInstance, icReadWritePerm);
		if err = noErr then begin

			(* Some special hacks for the Fonts window. *)
			dialogID := MapWindowTypeToResourceID(wt);
			if (wt = WT_Font) then begin
				(* Use the System 6 Fonts DLOG resource. Most of the time we use our
					own popup menu CDEF provided by Stuart.  However we got a lot of
					reports from people that the Fonts window would crash. So we
					switched the Fonts DLOG to use the standard system CDEF.
					However that CDEF isn't available under System 6, so we switch
					back to our own version under System 6.
				*)
				if not gHasSystem7 then begin
					dialogID := rFontsDialog6;
				end; (* if *)
				(* Switch to watch cursor, 'cause the Fonts dialog often takes a while. *)
				SetCursor(GetCursor(watchCursor)^^);	
			end; (* if *)
			
			(* Create the window and prepare the Whats. *)
			gWindows[wt].window := GetNewDialog(dialogID, nil, POINTER(-1));
			if gWindows[wt].window = nil then begin
				err := memFullErr;
			end; (* if *)
			if err = noErr then begin
				err := PrepareWindowWhats(wt);
			end; (* if *)

			(* Position the window. *)
			if err = noErr then begin
				(* Get the windows out of our global array. *)
				savedPosition := gWindows[wt].position;
				if (savedPosition.h <> 0) or (savedPosition.v <> 0) then begin
					(* There's useful data in the window positions, let's try to move the window there.
						But first, save away the position the window was created with, just in case
						the saved position proves to be offscreen.
					*)
					GetWindowRect(gWindows[wt].window, originalPosition);
					MoveWindow(gWindows[wt].window, savedPosition.h, savedPosition.v, false);
					(* We have to show the window now in this case, because TitleBarOnScreen 
						requires window to be shown and there's no workaround for this.
					*)
					ShowWindow(gWindows[wt].window);
					if not TitleBarOnScreen(gWindows[wt].window) then begin
						(* Hmmm, saved window position doesn't leave the title bar show. Move
							the window back to the original position.
						*)
						MoveWindow(gWindows[wt].window, originalPosition.left, originalPosition.top, false);
					end; (* if *)
				end; (* if *)
				ShowWindow(gWindows[wt].window);
				(* Unilaterally switch back to arrow cursor. *)
				InitCursor;
			end; (* if *)

			(* Clean up the IC session. *)
			err2 := ICEnd(GetDocumentInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		(* Clean up after failure. *)
		if err <> noErr then begin
			if gWindows[wt].window <> nil then begin
				junk := DisposeWindowType(wt);
			end; (* if *)
		end; (* if *)
		CreateNewWindow := err;
	end; (* CreateNewWindow *)

	(* ***** Opening and Closing ***** *)

	function WindowOpen (wt: WindowType): OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		err := noErr;
		case wt of
			WT_None :
				(* do nothing *) ;
			WT_About :
				begin
					(* The about box is created when we start up, so all we have to do is show it. *)
					ShowWindow(gWindows[wt].window);
					SelectWindow(gWindows[wt].window);
				end;
			otherwise
				begin
					if gWindows[wt].window <> nil then begin
						SelectWindow(gWindows[wt].window);
					end else begin
						err := CreateNewWindow(wt);
					end; (* if *)
				end
		end; (* case *)
		WindowOpen := err;
	end; (* WindowOpen *)

	function WindowClose (wt: WindowType): OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		err := noErr;
		case wt of
			WT_None :
				(* do nothing *) ;
			WT_About :
				HideWindow(gWindows[wt].window);
			otherwise
				err := CloseWindowType(wt);
		end; (* case *)
		WindowClose := err;
	end; (* WindowClose *)

	procedure WindowSetTitle (wt: WindowType; title: Str255);
		(* See comment in interface part. *)
	begin
		if gWindows[wt].window <> nil then begin
			SetWTitle(gWindows[wt].window, title);
		end; (* if *)
	end; (* WindowSetTitle *)

	function WindowsFlushAll: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			firstErr : OSStatus;
			wt: WindowType;
	begin
		(* Begin an IC session to speed up the operation of FlushWindowType. *)
		firstErr := ICBegin(GetDocumentInstance, icReadWritePerm);

		(* Flush each open information window. *)
		if firstErr = noErr then begin
			for wt := WT_Main to WT_Last do begin
				if gWindows[wt].window <> nil then begin
					SetPort(gWindows[wt].window);
					err := FlushWindowType(wt);
					if firstErr = noErr then begin
						firstErr := err;
					end; (* if *)
				end; (* if *)
			end; (* for *)
			
			(* Shut down the IC session and clean up. *)
			err := ICEnd(GetDocumentInstance);
			if firstErr = noErr then begin
				firstErr := err;
			end; (* if *)
		end; (* if *)
		WindowsFlushAll := firstErr;
	end; (* WindowsFlushAll *)

	function WindowsCloseAll: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			firstErr : OSStatus;
			wt: WindowType;
	begin
		firstErr := noErr;
		
		(* Close each open window. *)
		for wt := WT_Personal to WT_Last do begin
			if gWindows[wt].window <> nil then begin
				err := CloseWindowType(wt);
				if firstErr = noErr then begin
					firstErr := err;
				end; (* if *)
			end; (* if *)
		end; (* for *)
		
		WindowsCloseAll := firstErr;
	end; (* WindowsCloseAll *)

	(* ***** Window Positioning ***** *)

	type
		WindowPointArray = array[WT_Main..WT_Last] of Point;
		(* We have to declare our own type for saving the window positions because
			the gWindows array is full of other stuff we don't want to save. It has to be
			declared outside of the save and restore routines because it's shared between
			them.
		*)

	procedure WindowsResetPositions;
		(* See comment in interface part. *)
		var
			wt: WindowType;
			pos: Point;
			horizontalStagger : integer;
			verticalStagger : integer;
	begin
		(* Set up the initial position for the first window. *)
		pos.h := 2;
		pos.v := 42;
		
		(* Set up horizontal and vertical stagger values based on the size of the main monitor. *)
		horizontalStagger := 0;
		if qd.screenBits.bounds.right > 512 then begin
			horizontalStagger := horizontalStagger + 20;
		end; (* if *)
		verticalStagger := 0;
		if qd.screenBits.bounds.bottom >= 400 then begin
			verticalStagger := verticalStagger + 8;
		end; (* if *)
		if qd.screenBits.bounds.bottom >= 480 then begin
			verticalStagger := verticalStagger + 8;
		end; (* if *)
		if qd.screenBits.bounds.bottom > 480 then begin
			verticalStagger := verticalStagger + 2;
		end; (* if *)

		(* Now setup the default positions for the windows. *)
		for wt := WT_Main to pred(WT_Last) do begin
			gWindows[wt].position := pos;
			pos.h := pos.h + horizontalStagger;
			pos.v := pos.v + verticalStagger;
		end; (* for *)
	end; (* WindowsResetPositions *)

	procedure WindowsSavePositions;
		(* See comment in interface part. *)
		var
			err : OSStatus;
			err2: OSStatus;
			wt: WindowType;
			windowPositions: WindowPointArray;
	begin
		for wt := WT_Main to WT_Last do begin
			windowPositions[wt] := gWindows[wt].position;
		end; (* for *)
		err := ICBegin(GetDocumentInstance, icReadWritePerm);
		if err = noErr then begin
			err := ICSetPref(GetDocumentInstance, kOurWindowPositionsKey, ICattr_no_change, 
										@windowPositions, sizeof(windowPositions));
			err2 := ICEnd(GetDocumentInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		(* junk the error because we really don't care *)
	end; (* WindowsSavePositions *)

	procedure WindowsRestorePositions;
		(* See comment in interface part. *)
		var
			err : OSStatus;
			err2: OSStatus;
			junkAttr: longint;
			windowPositions: WindowPointArray;
			size: longint;
			wt: WindowType;
	begin
		err := ICBegin(GetDocumentInstance, icReadOnlyPerm);
		if err = noErr then begin
			size := sizeof(windowPositions);
			err := ICGetPref(GetDocumentInstance, kOurWindowPositionsKey, junkAttr, 
										@windowPositions, size);
			if (err = noErr) and (size <> sizeof(windowPositions)) then begin
				err := -1;
			end; (* if *)
			err2 := ICEnd(GetDocumentInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			for wt := WT_Main to WT_Last do begin
				gWindows[wt].position := windowPositions[wt];
			end; (* for *)
		end else begin
			WindowsResetPositions;
		end; (* if *)
	end; (* WindowsRestorePositions *)

	(* ***** Misc Stuff ***** *)

	function InitICWindows: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
	begin
		WindowsResetPositions;
		(* Bring the about box up hidden and leave it there. Can't call CreateNewWindow
			because it makes the assumption that the Whats want to have the IC
			file available and tries to call ICBegin, which doesn't work unless
			we have a document open, which we can't have until we've finished
			initialising this module!
		*)
		err := noErr;
		gWindows[WT_About].window := GetNewDialog(MapWindowTypeToResourceID(WT_About), nil, WindowPtr(-1));;
		if gWindows[WT_About].window = nil then begin
			err := memFullErr;
		end; (* if *)
		if err = noErr then begin
			err := PrepareWindowWhats(WT_About);
		end; (* if *)
		InitICWindows := err;
	end; (* InitICWindows *)

end. (* ICWindows *)
